from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from construct import (
    Adapter,
    Construct,
    MappingError,
)

if TYPE_CHECKING:
    from construct import Context

    TheAdapter = Adapter[bytes, bytes, "EnumByteString", str]
else:
    Context = "Context"  # pylint: disable=invalid-name
    TheAdapter = Adapter


class EnumByteString(str):
    """
    Like EnumIntegerString but for Bytes.
    """

    bytevalue: bytes

    def __repr__(self: Self) -> str:
        return f"EnumByteString.new({self.bytevalue!s}, {str.__repr__(self)})"

    def __int__(self: Self) -> bytes:
        return self.bytevalue

    @staticmethod
    def new(bytevalue: bytes, stringvalue: str) -> EnumByteString:
        """
        Create EnumByteString object as a class constructor
        """
        ret = EnumByteString(stringvalue)
        ret.bytevalue = bytevalue
        return ret


class EnumStr(TheAdapter):
    r"""
    This is exactly the same as construct.Enum, but supports byte subcons.
    """

    def __init__(self: Self, subcon: Construct[bytes, bytes], **mapping: bytes) -> None:
        super().__init__(subcon)
        self.encmapping: dict[str, bytes] = {
            EnumByteString.new(v, k): v for k, v in mapping.items()
        }
        self.decmapping: dict[bytes, EnumByteString] = {
            v: EnumByteString.new(v, k) for k, v in mapping.items()
        }
        self.ksymapping: dict[bytes, str] = {v: k for k, v in mapping.items()}

    def __getattr__(self, name: str) -> EnumByteString:
        if name in self.encmapping:
            return self.decmapping[self.encmapping[name]]
        raise AttributeError

    def _decode(self, obj: bytes, _: Context, __: str):  # type: ignore[no-untyped-def]
        try:
            return self.decmapping[obj]
        except KeyError:
            return EnumByteString.new(obj, "")

    def _encode(self, obj: str, _: Context, path: str):  # type: ignore[no-untyped-def]
        try:
            return self.encmapping[obj]
        except KeyError as exc:
            msg = f"building failed, no mapping for {obj!r}"
            raise MappingError(msg, path=path) from exc

    def _emitparse(self, code):  # type: ignore[no-untyped-def]
        fname = f"factory_{code.allocateId()}"
        code.append(f"{fname} = {self.decmapping!r}")
        return f"reuse(({self.subcon._compileparse(code)}), lambda x: {fname}.get(x, EnumInteger(x)))"  # type: ignore[attr-defined]  # pylint: disable=protected-access

    def _emitbuild(self, code):  # type: ignore[no-untyped-def]
        fname = f"factory_{code.allocateId()}"
        code.append(f"{fname} = {self.encmapping!r}")
        return f"reuse({fname}.get(obj, obj), lambda obj: ({self.subcon._compilebuild(code)}))"  # type: ignore[attr-defined]  # pylint: disable=protected-access

    def _emitprimitivetype(self, ksy, _):  # type: ignore[no-untyped-def]
        name = f"enum_{ksy.allocateId()}"
        ksy.enums[name] = self.ksymapping
        return name



from __future__ import annotations

from construct import (
    Bytes,
    Const,
    Struct,
    Switch,
)

from itksn.common import EnumStr

pcb_manufacturer = EnumStr(
    Bytes(1),
    EPEC=b"1",
    NCAB_100um=b"2",
    ATLAFLEX=b"3",
    SFCircuits=b"4",
    PHOENIX=b"5",
    Yamashita_Material=b"6",
    NCAB_75um=b"7",
    Tecnomec=b"8",
)

fe_chip_version = EnumStr(
    Bytes(1),
    RD53A=b"0",
    ITkpix_v1=b"1",
    ITkpix_v1p1=b"2",
    ITkpix_v2=b"3",
)

fe_chip_wafer = Struct(
    "batch_number"
    / EnumStr(
        Bytes(1),
        RD53A=b"0",
        ITkpix_v1=b"1",
        ITkpix_v2=b"2",
    ),
    "number" / Bytes(6),
)

sensor = Struct(
    "manufacturer"
    / EnumStr(Bytes(1), V1=b"0", V2=b"1", V3=b"2", V4=b"3", V5=b"4", V6=b"5"),
    "sensor_type"
    / EnumStr(
        Bytes(1),
        Single_RD53A=b"0",
        Single_ITkpix_v1_2=b"1",
        Double=b"2",
        Quad=b"3",
        Test_Structure_1=b"4",
        Test_Structure_2=b"5",
        Test_Structure_3=b"6",
        Test_Structure_4=b"7",
        Test_Structure_5=b"8",
        Test_Structure_6=b"9",
    ),
    "number" / Bytes(5),
)

bare_module = Struct(
    "FE_chip_version" / fe_chip_version,
    "sensor_type"
    / EnumStr(
        Bytes(1),
        No_sensor=b"0",
        Market_survey_sensor_tile=b"1",
        L0_inner_pixel_3D_sensor_tile=b"2",
        L0_inner_pixel_planar_sensor_tile=b"3",
        L1_inner_pixel_quad_sensor_tile=b"4",
        Outer_pixel_quad_sensor_tile=b"5",
    ),
    "number" / Bytes(5),
)

pcb = Struct(
    "FE_chip_version" / fe_chip_version,
    "PCB_manufacturer" / pcb_manufacturer,
    "number" / Bytes(5),
)

module_rd53a = Struct(
    "FE_chip_version" / fe_chip_version,
    "reserved" / Const(b"0"),
    "number" / Bytes(5),
)

module = Struct(
    "FE_chip_version" / fe_chip_version,
    "PCB_manufacturer" / pcb_manufacturer,
    "number" / Bytes(5),
)

module_carrier = Struct(
    "module_type"
    / EnumStr(
        Bytes(1),
        Quad_module_carrier=b"0",
        Cell_loaded_quad_module_bottom_cover=b"1",
        Linear_triplet_module_carrier=b"2",
        Ring_triplet_module_carrier=b"3",
    ),
    "module_version"
    / EnumStr(
        Bytes(1),
        Triplet_v1p0=b"1",
        Quad_v2p1=b"2",
    ),
    "manufacturer" / Bytes(1),
    "number" / Bytes(4),
)

subproject_codes = EnumStr(
    Bytes(2),
    FE_chip_wafer=b"FW",
    FE_chip=b"FC",
    Market_Survey_sensor_wafer=b"WM",
    L0_inner_pixel_3D_sensor_wafer=b"W0",
    L0_inner_pixel_planar_sensor_wafer=b"W1",
    L1_inner_pixel_sensor_wafer_thickness_100mum=b"W2",
    Outer_pixel_sensor_wafer_thickness_150mum=b"W3",
    Dummy_sensor_wafer=b"WD",
    Market_survey_sensor_tile=b"SM",
    L0_inner_pixel_3D_sensor_tile=b"S0",
    L0_inner_pixel_planar_sensor_tile=b"S1",
    L1_inner_pixel_quad_sensor_tile=b"S2",
    Outer_pixel_quad_sensor_tile=b"S3",
    Single_bare_module=b"B1",
    Dual_bare_module=b"B2",
    Quad_bare_module=b"B4",
    Digital_single_bare_module=b"BS",
    Digital_quad_bare_module=b"BQ",
    Dummy_single_bare_module=b"BT",
    Dummy_quad_bare_module=b"BR",
    Triplet_L0_Stave_PCB=b"PT",
    Triplet_L0_R0_PCB=b"P0",
    Triplet_L0_R0p5_PCB=b"P5",
    Quad_PCB=b"PQ",
    Dual_PCB=b"PD",
    PCB_test_coupon=b"PC",
    Triplet_L0_stave_module=b"MS",
    Triplet_L0_Ring0_module=b"M0",
    Triplet_L0_Ring0p5_module=b"M5",
    L1_quad_module=b"M1",
    Outer_system_quad_module=b"M2",
    Dual_chip_module=b"R2",
    Digital_triplet_module=b"R3",
    Digital_quad_module=b"R4",
    Dummy_triplet_module=b"R0",
    Dummy_quad_module=b"R1",
    Module_carrier=b"MC",
)

identifiers = Switch(
    lambda ctx: ctx.subproject_code,
    {
        "FE_chip_wafer": fe_chip_wafer,
        "Market_Survey_sensor_wafer": sensor,
        "L0_inner_pixel_3D_sensor_wafer": sensor,
        "L0_inner_pixel_planar_sensor_wafer": sensor,
        "L1_inner_pixel_sensor_wafer_thickness_100mum": sensor,
        "Outer_pixel_sensor_wafer_thickness_150mum": sensor,
        "Dummy_sensor_wafer": sensor,
        "Market_survey_sensor_tile": sensor,
        "L0_inner_pixel_3D_sensor_tile": sensor,
        "L0_inner_pixel_planar_sensor_tile": sensor,
        "L1_inner_pixel_quad_sensor_tile": sensor,
        "Outer_pixel_quad_sensor_tile": sensor,
        "Single_bare_module": bare_module,
        "Dual_bare_module": bare_module,
        "Quad_bare_module": bare_module,
        "Digital_single_bare_module": bare_module,
        "Digital_quad_bare_module": bare_module,
        "Dummy_single_bare_module": bare_module,
        "Dummy_quad_bare_module": bare_module,
        "Triplet_L0_Stave_PCB": pcb,
        "Triplet_L0_R0_PCB": pcb,
        "Triplet_L0_R0p5_PCB": pcb,
        "Quad_PCB": pcb,
        "Dual_PCB": pcb,
        "PCB_test_coupon": pcb,
        "Triplet_L0_stave_module": module_rd53a,
        "Triplet_L0_Ring0_module": module_rd53a,
        "Triplet_L0_Ring0p5_module": module_rd53a,
        "L1_quad_module": module,
        "Outer_system_quad_module": module,
        "Dual_chip_module": module_rd53a,
        "Digital_triplet_module": module_rd53a,
        "Digital_quad_module": module_rd53a,
        "Dummy_triplet_module": module_rd53a,
        "Dummy_quad_module": module_rd53a,
        "Module_carrier": module_carrier,
    },
    default=Bytes(7),
)



from __future__ import annotations

import typer

from itksn import __version__
from itksn.core import SerialNumberStruct

app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})


@app.callback(invoke_without_command=True)
def main(
    version: bool = typer.Option(False, "--version", help="Print the current version."),
) -> None:
    """
    Manage top-level options
    """
    if version:
        typer.echo(__version__)
        raise typer.Exit()


@app.command()
def parse(serialnumber: str) -> None:
    """
    Parse the provided serial number.
    """
    typer.echo(SerialNumberStruct.parse(serialnumber.encode("utf-8")))


# for generating documentation using mkdocs-click
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    typer.run(main)



from __future__ import annotations

import pytest
from construct import Bytes
from construct.core import TerminatedError

import itksn
from itksn.core import EnumStr


def test_enumstr():
    myenum = EnumStr(Bytes(2), itsaa=b"aa", itsbb=b"bb", itsyy=b"yy")
    assert myenum.parse(b"yy").bytevalue == b"yy"
    assert myenum.parse(b"yy") == "itsyy"
    assert myenum.parse(b"xx").bytevalue == b"xx"
    assert not myenum.parse(b"xx")


def test_parse_example1():
    parsed = itksn.parse(b"20Uxxyynnnnnnn")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert not parsed.project_code
    assert parsed.subproject_code == "yy"
    assert parsed.identifier == b"nnnnnnn"


def test_parse_example2():
    parsed = itksn.parse(b"20UPIyynnnnnnn")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "inner_pixel"
    assert not parsed.subproject_code
    assert parsed.identifier == b"nnnnnnn"


def test_parse_fe_wafer():
    parsed = itksn.parse(b"20UPGFW2123456")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "FE_chip_wafer"
    assert parsed.identifier.batch_number == "ITkpix_v2"
    assert parsed.identifier.number == b"123456"


def test_parse_sensor():
    parsed = itksn.parse(b"20UPGW34212345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "Outer_pixel_sensor_wafer_thickness_150mum"
    assert parsed.identifier.manufacturer == "V5"
    assert parsed.identifier.sensor_type == "Double"
    assert parsed.identifier.number == b"12345"


def test_parse_bare_module():
    parsed = itksn.parse(b"20UPGBS1112345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "Digital_single_bare_module"
    assert parsed.identifier.FE_chip_version == "ITkpix_v1"
    assert parsed.identifier.sensor_type == "Market_survey_sensor_tile"
    assert parsed.identifier.number == b"12345"


def test_parse_pcb():
    parsed = itksn.parse(b"20UPGPD0112345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "Dual_PCB"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.PCB_manufacturer == "EPEC"
    assert parsed.identifier.number == b"12345"


def test_parse_module():
    parsed = itksn.parse(b"20UPGR40012345")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "Digital_quad_module"
    assert parsed.identifier.FE_chip_version == "RD53A"
    assert parsed.identifier.reserved == b"0"
    assert parsed.identifier.number == b"12345"


def test_parse_quad_module():
    parsed = itksn.parse(b"20UPIM12602173")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "inner_pixel"
    assert parsed.subproject_code == "L1_quad_module"
    assert parsed.identifier.FE_chip_version == "ITkpix_v1p1"
    assert parsed.identifier.PCB_manufacturer == "Yamashita_Material"
    assert parsed.identifier.number == b"02173"


def test_parse_module_carrier():
    parsed = itksn.parse(b"20UPGMC2291234")
    assert parsed.atlas_project == "atlas_detector"
    assert parsed.system_code == "phaseII_upgrade"
    assert parsed.project_code == "pixel_general"
    assert parsed.subproject_code == "Module_carrier"
    assert parsed.identifier.module_type == "Linear_triplet_module_carrier"
    assert parsed.identifier.module_version == "Quad_v2p1"
    assert parsed.identifier.manufacturer == b"9"
    assert parsed.identifier.number == b"1234"


def test_parse_toomany():
    with pytest.raises(TerminatedError):
        itksn.parse(b"20Uxxyynnnnnnnmmmm")



from __future__ import annotations

import shlex
import time

import itksn


def test_version(script_runner):
    command = "itksn --version"
    start = time.time()
    ret = script_runner.run(shlex.split(command))
    end = time.time()
    elapsed = end - start
    assert ret.success
    assert itksn.__version__ in ret.stdout
    assert not ret.stderr
    # make sure it took less than 1.5s
    assert elapsed < 1.5


def test_parse(script_runner):
    command = "itksn parse 20Uxxyynnnnnnn"
    ret = script_runner.run(shlex.split(command))
    assert ret.success



from __future__ import annotations

from construct import (
    Bytes,
    PaddedString,
    Struct,
    Switch,
    Terminated,
    this,
)

from itksn import pixels
from itksn.common import EnumStr

SerialNumberStruct = "SerialNumber" / Struct(
    "atlas_project" / EnumStr(Bytes(2), atlas_detector=b"20"),
    "system_code"
    / EnumStr(
        Bytes(1),
        phaseII_upgrade=b"U",
    ),
    "project_code"
    / EnumStr(
        Bytes(2),
        inner_pixel=b"PI",
        outer_pixel_barrel=b"PB",
        pixel_general=b"PG",
        pixel_endcaps=b"PE",
        strip_general=b"SG",
        strip_barrel=b"SB",
        strip_endcaps=b"SE",
        common_mechanics=b"CM",
        common_electronics=b"CE",
    ),
    "subproject_code"
    / Switch(
        this.project_code,
        {
            "inner_pixel": pixels.subproject_codes,
            "outer_pixel_barrel": pixels.subproject_codes,
            "pixel_general": pixels.subproject_codes,
        },
        default=PaddedString(2, "utf8"),
    ),
    "identifier"
    / Switch(
        this.project_code,
        {
            "inner_pixel": pixels.identifiers,
            "outer_pixel_barrel": pixels.identifiers,
            "pixel_general": pixels.identifiers,
        },
        default=Bytes(7),
    ),
    Terminated,
)
